// Copyright 2025 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
//
// SPDX-License-Identifier: GPL-2.0-or-later
use lazy_static::lazy_static;
use regex::bytes::RegexSet;
use serde_json::Value;
use std::cmp::Ordering;
use std::collections::HashMap;
use tracing::info;

use hg::repo::Repo;
use hg::revlog::{changelog::Changelog, NodePrefix, RevlogError};

use askalono::Store;

use crate::gitaly::FindLicenseResponse;
use crate::mercurial::{changelog_entry_manifest, read_blob};

/// This is the same value as in the askalono CLI program.
///
/// For the record, go-license-detector has a `similarityThreshold` constant
/// but it uses a different simiarity algorithm
const ASKALONO_MIN_SCORE: f32 = 0.8;

lazy_static! {
    static ref LICENSE_FILE_NAMES_RX: RegexSet = RegexSet::new([
        "(?i-u)^li[cs]en[cs]e(s?)$",
        "(?i-u)^legal$",
        "(?i-u)^copy(left|right|ing)$",
        "(?i-u)^unlicense$",
        "(?i-u)^l?gpl([-_ v]?)(\\d\\.?\\d)?$",
        "(?i-u)^bsd$",
        "(?i-u)^mit$",
        "(?i-u)^apache$",
    ])
    .expect("Regex should compile");
}

lazy_static! {
    static ref NICKNAMES: HashMap<String, String> = {
        let mut nicks = HashMap::new();
        let raw = include_str!("license_nicknames.json");
        if let Ok(Some(Value::Object(map))) = serde_json::from_str(raw) {
            for (id, nick_val) in map.into_iter() {
                if let Value::String(nick) = nick_val {
                    nicks.insert(id, nick.to_owned());
                }
            }
        } else {
            panic!("License nicknames is not a mapping");
        }
        nicks
    };
}

/// Load nicknames from JSON
///
/// Useful to break early, at startup rather than potentially hours after
/// deployment when `license_analysis` is called for the first time.
pub fn load_license_nicknames() {
    assert!(NICKNAMES.len() > 0); // not entirely trivial to avoid being optimized away
}

pub fn license_analysis(
    licenses_db: &Store,
    repo: &Repo,
    changelog: &Changelog,
    node_prefix: NodePrefix,
) -> Result<FindLicenseResponse, RevlogError> {
    let manifestlog = repo.manifestlog()?;
    if let Some((node, manifest)) = changelog_entry_manifest(changelog, &manifestlog, node_prefix)?
    {
        let mut matches = Vec::new();
        for entry in manifest.iter() {
            let entry = entry?;
            let hg_path = entry.path;
            // TODO go-licenses-detector (investigation.go) would also:
            //
            // - handle the case where the license file contains a path to the actual license
            //   (trheshold of 128 bytes to detect this)
            // - look into a directory likely to hold licenses (with a separate regexp of its own)
            // - inspect README files (with a range of possible extensions)
            if LICENSE_FILE_NAMES_RX.is_match(hg_path.as_bytes()) {
                info!(
                    "Potential License file {}",
                    String::from_utf8_lossy(hg_path.as_bytes())
                );
                let (data, _metadata) = read_blob(repo, &node, entry, hg_path)?;
                let data = askalono::TextData::new(&String::from_utf8_lossy(&data));
                let m = licenses_db.analyze(&data);
                matches.push((m, String::from_utf8_lossy(hg_path.as_bytes()).to_string()));
            }
        }
        matches.sort_by(|(m1, _p1), (m2, _p2)| {
            // ordering by reverse score, and name in case of equality
            if m1.score == m2.score {
                m1.name.cmp(m2.name)
            } else if m1.score < m2.score {
                // No we won't have a NaN
                Ordering::Greater
            } else {
                Ordering::Less
            }
        });
        let best_match = matches.into_iter().next();
        if let Some((m, path)) = best_match {
            let details = spdx::license_id(m.name).unwrap();
            if m.score < ASKALONO_MIN_SCORE {
                Ok(FindLicenseResponse {
                    license_short_name: "other".to_string(),
                    license_name: "Other".to_string(),
                    license_path: path,
                    license_nickname: "LICENSE".to_string(),
                    ..Default::default()
                })
            } else {
                let lower_name = m.name.to_lowercase();
                let nickname = NICKNAMES
                    .get(&lower_name)
                    .map_or_else(String::new, |s| s.to_string());
                Ok(FindLicenseResponse {
                    license_short_name: lower_name,
                    license_name: details.full_name.to_string(),
                    license_path: path,
                    license_url: format!("https://spdx.org/licenses/{}.html", m.name),
                    license_nickname: nickname,
                })
            }
        } else {
            // no good enough match
            Ok(FindLicenseResponse::default())
        }
    } else {
        // called on NULL_REVISION
        Ok(FindLicenseResponse::default())
    }
}
